import React, { useCallback, useEffect, useMemo, useRef, useState } from "react";
import '../../../css/endo/EndoGrid.css';
import '../../../css/sitewide/colors.css';
import { MAX_ROWS } from "../Mito";
import { TaskpaneType } from "../taskpanes/taskpanes";
import CellEditor from "./CellEditor";
import { getCellEditorInputCurrentSelection, getStartingFormula } from "./cellEditorUtils";
import ColumnHeaders from "./ColumnHeaders";
import EmptyGridMessages from "./EmptyGridMessages";
import { focusGrid } from "./focusUtils";
import GridData from "./GridData";
import IndexHeaders from "./IndexHeaders";
import { equalSelections, getColumnIndexesInSelection, getIndexesFromMouseEvent, getIsHeader, getNewSelectionAfterKeyPress, getNewSelectionAfterMouseUp, isNavigationKeyPressed, isSelectionOnlyColumnHeaders, reconciliateSelection } from "./selectionUtils";
import { calculateCurrentSheetView, calculateNewScrollPosition, calculateTranslate, columnIsVisible, rowIsVisible } from "./sheetViewUtils";
import { firstNonNullOrUndefined, isNumberInRangeInclusive } from "./utils";
import { reconciliateWidthData } from "./widthUtils";
// NOTE: these should match the css
export const DEFAULT_WIDTH = 123;
export const DEFAULT_HEIGHT = 25;
export const MIN_WIDTH = 50;
export const KEYS_TO_IGNORE_IF_PRESSED_ALONE = [
    'Shift',
    'Meta',
    'Alt',
    'Control',
    'CapsLock',
    'Unidentified' // If you press the fn key on windows, this is the key
];
function EndoGrid(props) {
    // The container for the entire EndoGrid
    const containerRef = useRef(null);
    // The container for just the empty scroll div, and the rendered grid data
    const scrollAndRenderedContainerRef = useRef(null);
    // Store if the mouse is currently pressed down on the grid
    const [mouseDown, setMouseDown] = useState(false);
    // Destructure the props, so we access them more directly in the component below
    const { sheetData, sheetIndex, gridState, setGridState, cellEditorState, setCellEditorState, mitoAPI, setUIState } = props;
    const totalSize = {
        width: gridState.widthData.totalWidth,
        height: DEFAULT_HEIGHT * Math.min(((sheetData === null || sheetData === void 0 ? void 0 : sheetData.numRows) || 0), MAX_ROWS) // Limit the total height, as we display at most 2k rows, no matter what the df size is
    };
    const currentSheetView = useMemo(() => {
        return calculateCurrentSheetView(gridState);
    }, [gridState]);
    const translate = useMemo(() => {
        return calculateTranslate(gridState);
    }, [gridState]);
    /*
        An effect that handles the sheet data changing, in which case
        we have to perform a reconciliation of width data, as well
        as the selection

        Columns may have been deleted or added. We need to make sure that
        the widths and selection track these changes correctly.
    */
    useEffect(() => {
        setGridState(gridState => {
            return Object.assign(Object.assign({}, gridState), { selection: reconciliateSelection(gridState.sheetIndex, sheetIndex, gridState.selection, gridState.columnIDsArray, sheetData), widthData: reconciliateWidthData(gridState.sheetIndex, sheetIndex, gridState.widthData, gridState.columnIDsArray, sheetData), columnIDsArray: (sheetData === null || sheetData === void 0 ? void 0 : sheetData.data.map(c => c.columnID)) || [], sheetIndex: sheetIndex });
        });
    }, [sheetData, setGridState, sheetIndex]);
    /*
        An effect that handles a resizing of the viewport. Notably, the resize
        event is only triggered on the window event, so we cannot just add
        a resize event lister to the grid element and expect it to work.
        See here: https://developer.mozilla.org/en-US/docs/Web/API/Window/resize_event

        Thus, we use a resize observer.
        See here: https://developer.mozilla.org/en-US/docs/Web/API/Resize_Observer_API
    */
    useEffect(() => {
        const resizeViewport = () => {
            setGridState((gridState) => {
                var _a, _b;
                return Object.assign(Object.assign({}, gridState), { viewport: {
                        width: ((_a = scrollAndRenderedContainerRef === null || scrollAndRenderedContainerRef === void 0 ? void 0 : scrollAndRenderedContainerRef.current) === null || _a === void 0 ? void 0 : _a.clientWidth) || 0,
                        height: ((_b = scrollAndRenderedContainerRef === null || scrollAndRenderedContainerRef === void 0 ? void 0 : scrollAndRenderedContainerRef.current) === null || _b === void 0 ? void 0 : _b.clientHeight) || 0,
                    } });
            });
        };
        // Double calc the viewport size, just to make sure it loads properly
        resizeViewport();
        setTimeout(() => resizeViewport(), 250);
        const resizeObserver = new ResizeObserver(() => {
            resizeViewport();
        });
        const containerDiv = containerRef.current;
        if (containerDiv) {
            resizeObserver.observe(containerDiv);
        }
        return () => {
            resizeObserver.disconnect();
        };
    }, [setGridState]);
    // Handles a scroll inside the grid 
    const onGridScroll = (e) => {
        const newScrollPosition = calculateNewScrollPosition(e, totalSize, gridState.viewport, scrollAndRenderedContainerRef.current);
        if (newScrollPosition !== undefined) {
            setGridState((gridState) => {
                return Object.assign(Object.assign({}, gridState), { scrollPosition: newScrollPosition });
            });
        }
    };
    const onMouseDown = (e) => {
        var _a, _b;
        if (cellEditorState !== undefined) {
            // EDITING MODE
            const { rowIndex, columnIndex } = getIndexesFromMouseEvent(e);
            if (columnIndex !== undefined && (sheetData === null || sheetData === void 0 ? void 0 : sheetData.data[columnIndex]) !== undefined) {
                // Get the column that was clicked, and then find the current selection
                // within the cell editor, so that we can place the column header correctly
                const { selectionStart, selectionEnd } = getCellEditorInputCurrentSelection(containerRef.current);
                // If there is already some suggested column headers, we do not change this selection, 
                // as we want any future expanded selection of column headers to overwrite the same 
                // region. So default to pendingSelectedColumns?.selectionStart, but if this does not
                // exist, than take the selection range in the input currently
                const newSelectionStart = firstNonNullOrUndefined((_a = cellEditorState.pendingSelectedColumns) === null || _a === void 0 ? void 0 : _a.selectionStart, selectionStart);
                const newSelectionEnd = firstNonNullOrUndefined((_b = cellEditorState.pendingSelectedColumns) === null || _b === void 0 ? void 0 : _b.selectionEnd, selectionEnd);
                // Select the column that was clicked on, as they do in Excel
                setGridState(prevGridState => {
                    return Object.assign(Object.assign({}, prevGridState), { selection: {
                            startingRowIndex: rowIndex !== undefined ? rowIndex : -1,
                            endingRowIndex: rowIndex !== undefined ? rowIndex : -1,
                            startingColumnIndex: columnIndex,
                            endingColumnIndex: columnIndex,
                        } });
                });
                return setCellEditorState(Object.assign(Object.assign({}, cellEditorState), { pendingSelectedColumns: {
                        columnHeaders: [sheetData.data[columnIndex].columnHeader],
                        selectionStart: newSelectionStart,
                        selectionEnd: newSelectionEnd,
                    }, 
                    /* If you click on a cell, you should now scroll in the sheet */
                    arrowKeysScrollInFormula: false }));
            }
            return;
        }
        else {
            // NAVIGATION MODE
            // First, we make sure that the grid is focused on
            focusGrid(containerRef.current);
            // Return on right clicks, as the user is not actually interaction
            // with the grid, they are probably trying to inspect something
            if (e.metaKey) {
                return;
            }
            // Set state so we know mouse is down
            setMouseDown(true);
            // Update the selection
            const { rowIndex, columnIndex } = getIndexesFromMouseEvent(e);
            // If the click was not on a cell, return
            if (rowIndex === undefined || columnIndex === undefined) {
                return;
            }
            if (e.shiftKey) {
                // If the shift key is down, we extend the current selection
                setGridState((gridState) => {
                    return Object.assign(Object.assign({}, gridState), { selection: getNewSelectionAfterMouseUp(gridState.selection, rowIndex, columnIndex) });
                });
            }
            else {
                // Set the selection to be this single cell
                setGridState((gridState) => {
                    return Object.assign(Object.assign({}, gridState), { selection: {
                            startingRowIndex: rowIndex,
                            endingRowIndex: rowIndex,
                            startingColumnIndex: columnIndex,
                            endingColumnIndex: columnIndex,
                        } });
                });
            }
            // If the user is clicking on the index column, we make sure to close the control
            // panel if it's open, so we don't display something empty
            if (columnIndex === -1) {
                setUIState(prevUIState => {
                    if (prevUIState.currOpenTaskpane.type === TaskpaneType.CONTROL_PANEL) {
                        return Object.assign(Object.assign({}, prevUIState), { currOpenTaskpane: { type: TaskpaneType.NONE } });
                    }
                    return prevUIState;
                });
            }
        }
    };
    const onMouseUp = (e) => {
        // Do nothing, if we're in EDITING mode. Clicks handled by onMouseDown
        if (cellEditorState !== undefined) {
            return;
        }
        // Make sure the grid is focused
        focusGrid(containerRef.current);
        // Mark that the mouse is no longer down
        setMouseDown(false);
        const { rowIndex, columnIndex } = getIndexesFromMouseEvent(e);
        // If the shift key is down, then this is handled by the onMouseDown
        if (e.shiftKey) {
            return;
        }
        const newSelection = getNewSelectionAfterMouseUp(gridState.selection, rowIndex, columnIndex);
        // We only update the selection if has changed, so we don't rerender unnecessarily
        if (!equalSelections(newSelection, gridState.selection)) {
            setGridState((gridState) => {
                return Object.assign(Object.assign({}, gridState), { selection: newSelection });
            });
        }
    };
    // An effect so that when the mouse is down, the selection tracks where
    // the mouse is and updates live
    useEffect(() => {
        if (mouseDown) {
            const updateSelectionOnMouseDrag = (e) => {
                const { rowIndex, columnIndex } = getIndexesFromMouseEvent(e);
                setGridState((gridState) => {
                    return Object.assign(Object.assign({}, gridState), { selection: getNewSelectionAfterMouseUp(gridState.selection, rowIndex, columnIndex) });
                });
            };
            const containerDiv = containerRef.current;
            // We don't allow the drag and drop selections if you're starting from a column 
            // header, because the headers themselves are draggable and droppable
            if (gridState.selection.startingRowIndex === -1) {
                return;
            }
            containerDiv === null || containerDiv === void 0 ? void 0 : containerDiv.addEventListener('mousemove', updateSelectionOnMouseDrag);
            return () => {
                containerDiv === null || containerDiv === void 0 ? void 0 : containerDiv.removeEventListener('mousemove', updateSelectionOnMouseDrag);
            };
        }
    }, [mouseDown, gridState, setGridState]);
    // On double click, open the cell editor on this cell
    const onDoubleClick = (e) => {
        const { rowIndex, columnIndex } = getIndexesFromMouseEvent(e);
        // Don't open for headers
        if ((rowIndex === undefined || columnIndex === undefined) || getIsHeader(rowIndex, columnIndex)) {
            return;
        }
        const startingFormula = getStartingFormula(rowIndex, columnIndex, sheetData);
        setCellEditorState({
            rowIndex: rowIndex,
            columnIndex: columnIndex,
            formula: startingFormula,
            // As in google sheets, if the starting formula is non empty, we default to the 
            // arrow keys scrolling in the editor
            arrowKeysScrollInFormula: startingFormula.length > 0
        });
    };
    // A helper to scroll a given row into view
    const scrollRowIntoView = useCallback((rowIndex) => {
        // The column headers are always visible, so we don't have to do anything
        if (rowIndex === -1) {
            return;
        }
        const gridDiv = scrollAndRenderedContainerRef.current;
        if (gridDiv === null)
            return;
        let scrollTop = gridDiv.scrollTop;
        const rowVisible = rowIsVisible(containerRef.current, rowIndex);
        if (!rowVisible) {
            const newCellIsAbove = rowIndex <= currentSheetView.startingRowIndex;
            if (newCellIsAbove) {
                scrollTop = (rowIndex) * DEFAULT_HEIGHT;
            }
            else {
                scrollTop = (rowIndex + 1) * DEFAULT_HEIGHT - (gridDiv.clientHeight);
            }
        }
        gridDiv.scrollTop = scrollTop;
    }, [currentSheetView]);
    // A helper to scroll a column into view
    const scrollColumnIntoView = useCallback((columnIndex) => {
        // The index headers are always visible, so we don't have to do anything
        if (columnIndex === -1) {
            return;
        }
        const gridDiv = scrollAndRenderedContainerRef.current;
        if (gridDiv === null)
            return;
        let scrollLeft = gridDiv.scrollLeft;
        const columnVisible = columnIsVisible(containerRef.current, columnIndex);
        if (!columnVisible) {
            const newCellIsLeft = columnIndex <= currentSheetView.startingColumnIndex;
            if (newCellIsLeft) {
                scrollLeft = gridState.widthData.widthSumArray[columnIndex - 1];
            }
            else {
                // Put it at the start, move it to the end, then move it back one
                scrollLeft = gridState.widthData.widthSumArray[columnIndex] - (gridDiv.clientWidth);
            }
        }
        gridDiv.scrollLeft = scrollLeft;
    }, [currentSheetView, gridState.widthData]);
    /*
        Makes sure the given rowIndex and columnIndex are visible, by scrolling
        the screen the minimal amount to make them visible
    */
    const ensureCellVisible = useCallback((rowIndex, columnIndex) => {
        /*
            For some reason, there is an incredibly hard to find / track down bug where
            if you use the metaKey to scroll a huge number of cells at once, then the
            scrolls get screwed up. Specifically, the manifests as a bug where the
            scrollLeft/scrollTop ends up getting set when it shouldn't be, and we scroll
            to where we don't want.

            I think this is a race condition, and the simple fix is to just scroll the
            row or column after a small timeout in the case that we move the sheet view
            some large number of rows or columns.
        */
        const largeRowJump = !isNumberInRangeInclusive(rowIndex, currentSheetView.startingRowIndex - 2, currentSheetView.startingRowIndex + currentSheetView.numRowsRendered + 1);
        const largeColumnJump = !isNumberInRangeInclusive(columnIndex, currentSheetView.startingColumnIndex - 2, currentSheetView.startingColumnIndex + currentSheetView.numColumnsRendered + 1);
        const largeJump = largeRowJump || largeColumnJump;
        // We save these both before, to try and avoid race conditions
        const rowVisible = rowIsVisible(containerRef.current, rowIndex);
        const columnVisible = columnIsVisible(containerRef.current, columnIndex);
        // Make the row visible
        if (!rowVisible) {
            if (!largeJump) {
                scrollRowIntoView(rowIndex);
            }
            else {
                setTimeout(() => scrollRowIntoView(rowIndex), 25);
            }
        }
        // Make the column visible
        if (!columnVisible) {
            if (!largeJump) {
                scrollColumnIntoView(columnIndex);
            }
            else {
                setTimeout(() => scrollColumnIntoView(columnIndex), 50);
            }
        }
    }, [currentSheetView, scrollRowIntoView, scrollColumnIntoView]);
    // Effect listeners for when keys are pressed
    useEffect(() => {
        const onKeyDown = (e) => {
            var _a;
            // If we're in editing mode, then we don't do anything with the keyboard 
            // events as they are handeled by the cell editor state machine!
            if (cellEditorState !== undefined) {
                return;
            }
            if (KEYS_TO_IGNORE_IF_PRESSED_ALONE.includes(e.key)) {
                return;
            }
            if (!isNavigationKeyPressed(e.key)) {
                // If the metaKey is pressed, the user might be refreshing the page for example, 
                // so we just return here
                if (e.metaKey || e.key === 'Escape') {
                    return;
                }
                // If the key pressed is delete, and the user is selecting some column headers,
                // then we delete the columns they have selected
                if (e.key === 'Backspace' && isSelectionOnlyColumnHeaders(gridState.selection)) {
                    const columnIndexesSelected = getColumnIndexesInSelection(gridState.selection);
                    // TODO make this delete all columns
                    const columnIDToDelete = (_a = sheetData === null || sheetData === void 0 ? void 0 : sheetData.data[columnIndexesSelected[0]]) === null || _a === void 0 ? void 0 : _a.columnID;
                    if (columnIDToDelete !== undefined) {
                        void mitoAPI.sendDeleteColumn(sheetIndex, columnIDToDelete);
                    }
                }
                // If we press any key that is not a navigation key, then we open the editor
                setGridState((gridState) => {
                    const selection = gridState.selection;
                    // Since we cannot edit the column headers (yet!)
                    // TODO: in the future, we can allow users to edit column headers like this?
                    const isHeader = getIsHeader(selection.startingRowIndex, selection.startingColumnIndex);
                    if (isHeader) {
                        return Object.assign(Object.assign({}, gridState), { selection: selection });
                    }
                    const startingFormula = getStartingFormula(selection.startingRowIndex, selection.startingColumnIndex, sheetData, e);
                    setCellEditorState({
                        rowIndex: selection.startingRowIndex,
                        columnIndex: selection.startingColumnIndex,
                        formula: startingFormula,
                        // As in google sheets, if the starting formula is non empty, we default to the 
                        // arrow keys scrolling in the editor
                        arrowKeysScrollInFormula: startingFormula.length > 0
                    });
                    e.preventDefault();
                    return Object.assign(Object.assign({}, gridState), { selection: {
                            startingRowIndex: selection.startingRowIndex,
                            endingRowIndex: selection.startingRowIndex,
                            startingColumnIndex: selection.startingColumnIndex,
                            endingColumnIndex: selection.startingColumnIndex
                        } });
                });
                return;
            }
            else {
                // Otherwise, a navigation key was pressed, and so we should navigate!
                // Prevent the default of the key (to scroll or tab)
                e.preventDefault();
                // Update the selection
                setGridState((gridState) => {
                    const newSelection = getNewSelectionAfterKeyPress(gridState.selection, e, sheetData);
                    ensureCellVisible(newSelection.endingRowIndex, newSelection.endingColumnIndex);
                    return Object.assign(Object.assign({}, gridState), { selection: newSelection });
                });
            }
        };
        const containerDiv = containerRef.current;
        containerDiv === null || containerDiv === void 0 ? void 0 : containerDiv.addEventListener('keydown', onKeyDown);
        return () => containerDiv === null || containerDiv === void 0 ? void 0 : containerDiv.removeEventListener('keydown', onKeyDown);
    }, [cellEditorState, setCellEditorState, sheetData, currentSheetView, ensureCellVisible, mitoAPI, gridState.selection, sheetIndex, setGridState]);
    return (React.createElement("div", { className: 'endo-grid-container', ref: containerRef, tabIndex: -1, onMouseDown: onMouseDown, onMouseUp: onMouseUp, onMouseLeave: () => setMouseDown(false), onDoubleClick: onDoubleClick },
        sheetData !== undefined &&
            React.createElement(React.Fragment, null,
                React.createElement(ColumnHeaders, { sheetData: sheetData, setUIState: setUIState, sheetIndex: sheetIndex, containerRef: containerRef, scrollAndRenderedContainerRef: scrollAndRenderedContainerRef, gridState: gridState, setGridState: setGridState, mitoAPI: mitoAPI }),
                React.createElement(IndexHeaders, { sheetData: sheetData, gridState: gridState })),
        React.createElement("div", { className: "endo-scroller-and-renderer-container", ref: scrollAndRenderedContainerRef, onScroll: onGridScroll },
            React.createElement(EmptyGridMessages, { sheetData: sheetData }),
            React.createElement("div", { id: 'scroller', style: {
                    height: `${totalSize.height}px`,
                    width: `${totalSize.width}px`
                } }),
            React.createElement("div", { className: "renderer", style: {
                    transform: `translate(${gridState.scrollPosition.scrollLeft - translate.x}px, ${gridState.scrollPosition.scrollTop - translate.y}px)`,
                } },
                React.createElement(GridData, { sheetData: sheetData, gridState: gridState, cellEditorState: cellEditorState }))),
        sheetData !== undefined && cellEditorState !== undefined &&
            React.createElement(CellEditor, { sheetData: sheetData, sheetIndex: sheetIndex, gridState: gridState, cellEditorState: cellEditorState, setGridState: setGridState, setCellEditorState: setCellEditorState, scrollAndRenderedContainerRef: scrollAndRenderedContainerRef, containerRef: containerRef, ensureCellVisible: ensureCellVisible, mitoAPI: mitoAPI })));
}
export default EndoGrid;
//# sourceMappingURL=EndoGrid.js.map