import { DEFAULT_WIDTH } from "./EndoGrid";
/*
    A helper function for getting the default widthData
    based on some passed sheet data.

    Optionally, pass defaultWidthData which is a mapping
    from columnID -> width, to set the columns to some
    specific width.
*/
export const getWidthData = (sheetData, defaultWidthData = undefined) => {
    if (sheetData === undefined) {
        return {
            widthArray: [],
            widthSumArray: [],
            totalWidth: 0
        };
    }
    const widthArray = new Array(sheetData.numColumns);
    const widthSumArray = new Array(sheetData.numColumns);
    for (let columnIndex = 0; columnIndex < sheetData.numColumns; columnIndex++) {
        const columnID = sheetData.data[columnIndex].columnID;
        let columnWidth = DEFAULT_WIDTH;
        if (defaultWidthData !== undefined) {
            // If the defaultWidthData does not have this columnID stored, then take the default width
            columnWidth = defaultWidthData[columnID] || DEFAULT_WIDTH;
        }
        // NOTE that this is the width after this item!
        widthArray[columnIndex] = columnWidth;
        const previousSum = columnIndex === 0 ? 0 : widthSumArray[columnIndex - 1];
        widthSumArray[columnIndex] = previousSum + columnWidth;
    }
    return {
        widthArray: widthArray,
        widthSumArray: widthSumArray,
        totalWidth: widthArray.reduce((partialSum, width) => partialSum + width, 0)
    };
};
/*
    A helper function for changing with width of a specific
    column at a specific index.
*/
export const changeColumnWidth = (widthData, columnIndex, newWidth) => {
    // Update the width array
    const newWidthArray = [...widthData.widthArray];
    newWidthArray[columnIndex] = newWidth;
    const newWidthSumArray = [...widthData.widthSumArray];
    for (let i = columnIndex; i < widthData.widthSumArray.length; i++) {
        let previousSumSaved = newWidthSumArray[i - 1];
        if (previousSumSaved === undefined) {
            previousSumSaved = 0;
        }
        const columnWidth = newWidthArray[i];
        newWidthSumArray[i] = previousSumSaved + columnWidth;
    }
    // And finially the total width
    const newTotalWidth = newWidthArray.reduce((partialSum, width) => partialSum + width, 0);
    return {
        widthArray: newWidthArray,
        widthSumArray: newWidthSumArray,
        totalWidth: newTotalWidth
    };
};
/*
    Handles when sheet data changes, potentially reordering, adding, or removing
    columns.

    Specifically, it makes sure that the column widths are preserved even if the
    sheet data changes.

    TODO: this might be a performance hit in the future (but maybe not), so we might
    want to make this more efficient.
*/
export const reconciliateWidthData = (oldSheetIndex, newSheetIndex, widthData, oldColumnIDsArray, sheetData) => {
    if (oldSheetIndex !== newSheetIndex || sheetData === undefined) {
        return getWidthData(sheetData);
    }
    const oldWidths = {};
    for (let i = 0; i < oldColumnIDsArray.length; i++) {
        oldWidths[oldColumnIDsArray[i]] = widthData.widthArray[i];
    }
    return getWidthData(sheetData, oldWidths);
};
//# sourceMappingURL=widthUtils.js.map